

# C callback function for ref's (channels)
cdef void evproc_rslvstat(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                          void *info_ptr, void *privptr2) with gil:
    cdef BaseChan chan = <BaseChan>(<event*>privptr2).objptr
    chan.rslv_stat = <long>info_ptr # resolve status passed through this pointer
    chan.resolve.emit(chan)


cdef void evproc_update(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                        void *info_ptr, void *privptr2) with gil:
    cdef:
        cx_time_t timestr
        BaseChan chan = <BaseChan>(<event*>privptr2).objptr
    chan.check_exception( cda_get_ref_stat(ref, &chan.rflags, &timestr) )
    chan.prev_time = chan.time
    chan.time = <int64>timestr.sec * 1000000 + timestr.nsec / 1000
    chan.cb()


cdef void quant_update(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                        void *info_ptr, void *privptr2) with gil:
    cdef:
        int res
        CxAnyVal_t quant_raw
        cxdtype_t quant_dtype
        BaseChan chan = <BaseChan>(<event*>privptr2).objptr
        double quant_draw, val, shift

    res = cda_quant_of_ref(chan.ref, &quant_raw, &quant_dtype)
    quant_draw = aval_value(&quant_raw, quant_dtype)
    if quant_draw == 0:
        chan.quant = 0
        return
    else:
        res = cda_rd_convert(ref, 0, &shift)
        res = cda_rd_convert(ref, quant_draw, &val)
        chan.quant = val - shift


cdef void range_update(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                        void *info_ptr, void *privptr2) with gil:
    cdef:
        BaseChan chan = <BaseChan>(<event*>privptr2).objptr
    chan.get_range()

cdef void strs_update(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                        void *info_ptr, void *privptr2) with gil:
    cdef:
        BaseChan chan = <BaseChan>(<event*>privptr2).objptr
    chan.get_strings()


cdef void lockstat_update(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                        void *info_ptr, void *privptr2) with gil:
    cdef:
        BaseChan chan = <BaseChan>(<event*>privptr2).objptr
    chan.get_lockstate()


# wrapper-class for low-level functions and channel registration
cdef class BaseChan(CdaObject):
    cdef readonly:
        cda_dataref_t ref      # chan id
        str name               # chan name
        int max_nelems         # max number of elements when registered
        cxdtype_t dtype        # data type
        size_t itemsize        # item size in bytes
        int64 time, prev_time  # last and prev data update time
        long rslv_stat         # resolving status
        double quant
        int registered, first_cycle, lock_state
        Context context
        rflags_t rflags        # flags
        double rng[2]          # ranges
        # strings of ref
        char *ident
        char *label
        char *tip
        char *comment
        char *geoinfo
        char *rsrvd6
        char *units
        char *dpyfmt
        object valueMeasured, valueChanged, resolve, stringsUpdate, lockStateUpdate

    def __cinit__(self, str name, **kwargs):
        self.context = kwargs.get('context', default_context)
        self.max_nelems = kwargs.get('max_nelems', 1)

        # that's bad... purent class know about subs, but works to force dtypes fro known cases
        if isinstance(self, DChan) or isinstance(self, VPChan):
            self.dtype = CXDTYPE_DOUBLE
        elif isinstance(self, IChan):
            self.dtype = CXDTYPE_UINT32
        elif isinstance(self, StrChan):
            self.dtype = CXDTYPE_TEXT
        else:
            dtype = kwargs.get('dtype', CXDTYPE_DOUBLE)
            if type(dtype) == str:
                self.dtype = cx_dtype_map[dtype]
            else:
                self.dtype = dtype

        self.valueChanged, self.valueMeasured, self.resolve = Signal(object), Signal(object), Signal(object)
        self.stringsUpdate, self.lockStateUpdate = Signal(object), Signal(object)

        b_name = name.encode("ascii")
        cdef:
            char *c_name = b_name
            int ret
            unsigned int options = 0

        # may be cycle?
        if kwargs.get('private', False):
            options += CDA_DATAREF_OPT_PRIVATE

        if kwargs.get('no_rd_conv', False):
            options += CDA_DATAREF_OPT_NO_RD_CONV

        if kwargs.get('shy', False):
            options += CDA_DATAREF_OPT_SHY

        if kwargs.get('find_only', False):
            options += CDA_DATAREF_OPT_FIND_ONLY

        if kwargs.get('on_update', False):
            options += CDA_DATAREF_OPT_ON_UPDATE

        if kwargs.get('no_wr_wait', False):
            options += CDA_DATAREF_OPT_NO_WR_WAIT

        if kwargs.get('debug', False):
            options += CDA_DATAREF_OPT_DEBUG

        ret = cda_add_chan((<Context>self.context).cid, NULL, c_name, options, self.dtype, self.max_nelems,
                           0, <cda_dataref_evproc_t>NULL, NULL)
        self.check_exception(ret)
        self.ref, self.name, self.itemsize = ret, name, sizeof_cxdtype(self.dtype)

        (<Context>self.context).save_chan(<void*>self)

        # events which can be regular
        self.add_event(CDA_REF_EVMASK_RSLVSTAT, <void*>evproc_rslvstat, <void*>self, NULL)
        self.add_event(CDA_REF_EVMASK_UPDATE, <void*>evproc_update, <void*>self, NULL)

        self.add_event(CDA_REF_EVMASK_QUANTCHG, <void*>quant_update, <void*>self, NULL)
        self.add_event(CDA_REF_EVMASK_RANGECHG, <void*>range_update, <void*>self, NULL)

        self.add_event(CDA_REF_EVMASK_STRSCHG, <void*>strs_update, <void*>self, NULL)

        self.add_event(CDA_REF_EVMASK_LOCKSTAT, <void*>lockstat_update, <void*>self, NULL)

        if kwargs.get('get_curval', False):
            self.add_event(CDA_REF_EVMASK_CURVAL, <void*>evproc_update, <void*>self, NULL)

        # CX event masks
        ## <-- have some implementation
        # TODO: need to allow user to select registered events

        ## CDA_REF_EVMASK_UPDATE
        # CDA_REF_EVMASK_STATCHG
        ## CDA_REF_EVMASK_STRSCHG
        # CDA_REF_EVMASK_RDSCHG
        # CDA_REF_EVMASK_FRESHCHG
        ## CDA_REF_EVMASK_QUANTCHG
        ## CDA_REF_EVMASK_RANGECHG
        ## CDA_REF_EVMASK_RSLVSTAT
        ## CDA_REF_EVMASK_CURVAL
        ## CDA_REF_EVMASK_LOCKSTAT

        # initialization events (will be unregistered when get)

        self.registered, self.first_cycle = True, True
        self.rslv_stat = CDA_RSLVSTAT_SEARCHING
        self.rng[0],self.rng[1] = 0,0

    def __dealloc__(self):
        if self.registered:
            (<Context>self.context).drop_chan(<void*>self)
            cda_del_chan(self.ref)
            self.registered = 0

    def __str__(self):
        return '<cda_chan: ref=%d, name=%s>' % (self.ref, self.name)

    def short_name(self):
        a = self.name.split('.')
        return a[len(a)-1].split('@')[0]

    cdef void cb(self):
        #empty callback for overrideing
        pass

    cdef int snd_data(self, cxdtype_t dtype, int nelems, void* data_p):
        cdef int res = cda_snd_ref_data(self.ref, dtype, nelems, data_p)
        self.check_exception(res)

    cdef int get_data(self, size_t ofs, size_t size, void* buf):
        cdef int res = cda_get_ref_data(self.ref, ofs, size, buf)
        self.check_exception(res)
        return res

    cdef int current_nelems(self):
        return cda_current_nelems_of_ref(self.ref)

    cdef void get_src(self, const char **src_p):
        self.check_exception( cda_src_of_ref(self.ref, src_p) )

    cdef void register_event(self, event *ev):
        self.check_exception( cda_add_dataref_evproc(self.ref, ev.evmask, <cda_dataref_evproc_t>ev.evproc, ev) )

    cdef void unregister_event(self, event *ev):
        self.check_exception( cda_del_dataref_evproc(self.ref, ev.evmask, <cda_dataref_evproc_t>ev.evproc, ev) )

    cdef void check_exception(self, int c_res):
        if c_res < 0:
            raise Exception("cda chan error: cname=%s, %s, errcode=%s" % (self.name, cda_last_err(), c_res ))

    cpdef is_available(self):
        if self.rslv_stat == CDA_RSLVSTAT_FOUND:
            return True
        return False

    # TESTING functions, not yet fully implemented
    cpdef get_range(self):
        cdef:
            CxAnyVal_t r[2]
            double raw_r
            cxdtype_t dt
            int c_res

        c_res = cda_range_of_ref(self.ref, r, &dt)
        if dt == CXDTYPE_UNKNOWN:
            self.rng[0], self.rng[1] = 0, 0
        else:
            raw_r = aval_value(&(r[0]), dt)
            c_res = cda_rd_convert(self.ref, raw_r, &self.rng[0])

            raw_r = aval_value(&(r[1]), dt)
            c_res = cda_rd_convert(self.ref, raw_r, &self.rng[1])

    cpdef get_strings(self):
        c_res = cda_strings_of_ref(self.ref, &self.ident, &self.label, &self.tip, &self.comment,
                                   &self.geoinfo, &self.rsrvd6, &self.units, &self.dpyfmt)
        self.stringsUpdate.emit(self)

    cpdef get_lockstate(self):
        self.lock_state = cda_lock_stat_of_ref(self.ref)
        self.lockStateUpdate.emit(self)

    cpdef lock(self):
        cdef int ret = cda_lock_chans(1, &self.ref, CX_LOCK_WRITE_SET, 0)

    cpdef unlock(self):
        cdef int ret = cda_lock_chans(1, &self.ref, CX_LOCK_WRITE_RLS, 0)




# auxilary definitions and functions for CX events handling

# struct to extend private pointer
ctypedef struct event:
    int evmask    # cda eventmask
    void *evproc  # pointer to proc function
    void *objptr  # pointer to sender object
    void *userptr # pointer to user data





cpdef set rflags_text(rflags_t rflags):
    ret = set()
    if (rflags & CXRF_CAMAC_NO_X) > 0:
        ret.add('No X from CAMAC device')
    if (rflags & CXRF_CAMAC_NO_Q) > 0:
        ret.add('No Q from CAMAC device')
    if (rflags & CXRF_IO_TIMEOUT) > 0:
        ret.add('I/O timeout expired')
    if (rflags & CXRF_REM_C_PROBL) > 0:
        ret.add('Remote controller problem')
    if (rflags & CXRF_OVERLOAD) > 0:
        ret.add('Input channel overload')
    if (rflags & CXRF_UNSUPPORTED) > 0:
        ret.add('Unsupported feature/channel')
    if (rflags & CXRF_INVAL) > 0:
        ret.add('Invalid parameter')
    if (rflags & CXRF_WRONG_DEV) > 0:
        ret.add('Wrong device')
    if (rflags & CXRF_CFG_PROBL) > 0:
        ret.add('Configuration problem')
    if (rflags & CXRF_DRV_PROBL) > 0:
        ret.add('Driver internal problem')
    if (rflags & CXRF_NO_DRV) > 0:
        ret.add('Driver loading problem')
    if (rflags & CXRF_OFFLINE) > 0:
        ret.add('Device is offline')
    if (rflags & CXCF_FLAG_CALCERR) > 0:
        ret.add('Formula calculation error')
    if (rflags & CXCF_FLAG_DEFUNCT) > 0:
        ret.add('Defunct channel')
    if (rflags & CXCF_FLAG_OTHEROP) > 0:
        ret.add('Other operator is active')
    if (rflags & CXCF_FLAG_PRGLYCHG) > 0:
        ret.add('Channel was programmatically changed')
    if (rflags & CXCF_FLAG_NOTFOUND) > 0:
        ret.add('Channel not found')
    if (rflags & CXCF_FLAG_COLOR_WEIRD) > 0:
        ret.add('Value is weird')
    if (rflags & CXCF_FLAG_ALARM_ALARM) > 0:
        ret.add('Alarm!')
    if (rflags & CXCF_FLAG_ALARM_RELAX) > 0:
        ret.add('Relaxing after alarm')
    if (rflags & CXCF_FLAG_COLOR_RED) > 0:
        ret.add('Value in red zone')
    if (rflags & CXCF_FLAG_COLOR_YELLOW) > 0:
        ret.add('Value in yellow zone')
    return ret


cpdef tuple rflags_color_status(rflags_t rflags):
    if rflags == 0:
        return None, 'normal'
    elif (rflags & CXRF_CAMAC_NO_X) > 0 or\
         (rflags & CXRF_CAMAC_NO_Q) > 0 or\
         (rflags & CXRF_IO_TIMEOUT) > 0 or \
         (rflags & CXRF_REM_C_PROBL) > 0 or \
         (rflags & CXRF_OVERLOAD) > 0 or \
         (rflags & CXRF_UNSUPPORTED) > 0:
        return '#B03060', 'hwerr'
    elif (rflags & CXRF_INVAL) > 0 or\
         (rflags & CXRF_WRONG_DEV) > 0 or\
         (rflags & CXRF_CFG_PROBL) > 0 or\
         (rflags & CXRF_DRV_PROBL) > 0 or\
         (rflags & CXRF_NO_DRV) > 0 or \
         (rflags & CXRF_OFFLINE) > 0 or \
         (rflags & CXCF_FLAG_CALCERR) > 0:
        return '#8B8B00', 'sferr'
    elif (rflags & CXCF_FLAG_DEFUNCT) > 0:
        return '#4682B4', 'defunct'
    elif (rflags & CXCF_FLAG_OTHEROP) > 0:
        return '#FFA500', 'otherop'
    elif (rflags & CXCF_FLAG_PRGLYCHG) > 0:
        return '#D8E3D5', 'prglychg'
    elif (rflags & CXCF_FLAG_NOTFOUND) > 0:
        return '#404040', 'notfound'
    elif (rflags & CXCF_FLAG_COLOR_WEIRD) > 0:
        return '#0000FF', 'weird'
    elif (rflags & CXCF_FLAG_COLOR_YELLOW) > 0:
        return '#EDED6D', 'yellow'
    elif (rflags & CXCF_FLAG_COLOR_RED) > 0:
        return '#FFC0CB', 'red'
    elif (rflags & CXCF_FLAG_ALARM_ALARM) > 0:
        return '#FFC0CB', 'alarm'
    elif (rflags & CXCF_FLAG_ALARM_RELAX) > 0:
        return '#00FF00', 'alarm_relaxed'


rflags_order = [CXRF_CAMAC_NO_X,
               CXRF_CAMAC_NO_Q,
               CXRF_IO_TIMEOUT,
               CXRF_REM_C_PROBL,
               CXRF_OVERLOAD,
               CXRF_UNSUPPORTED,
               CXRF_INVAL,
               CXRF_WRONG_DEV,
               CXRF_CFG_PROBL,
               CXRF_DRV_PROBL,
               CXRF_NO_DRV,
               CXRF_OFFLINE,
               CXCF_FLAG_CALCERR,
               CXCF_FLAG_DEFUNCT,
               CXCF_FLAG_OTHEROP,
               CXCF_FLAG_PRGLYCHG,
               CXCF_FLAG_NOTFOUND,
               CXCF_FLAG_COLOR_WEIRD,
               CXCF_FLAG_ALARM_ALARM,
               CXCF_FLAG_ALARM_RELAX,
               CXCF_FLAG_COLOR_RED,
               CXCF_FLAG_COLOR_YELLOW
               ]

rflags_meaning = {
   CXRF_CAMAC_NO_X: 'No X from CAMAC device',
   CXRF_CAMAC_NO_Q: 'No Q from CAMAC device',
   CXRF_IO_TIMEOUT: 'I/O timeout expired',
   CXRF_REM_C_PROBL: 'Remote controller problem',
   CXRF_OVERLOAD: 'Input channel overload',
   CXRF_UNSUPPORTED: 'Unsupported feature/channel',
   CXRF_INVAL: 'Invalid parameter',
   CXRF_WRONG_DEV: 'Wrong device',
   CXRF_CFG_PROBL: 'Configuration problem',
   CXRF_DRV_PROBL: 'Driver internal problem',
   CXRF_NO_DRV: 'Driver loading problem',
   CXRF_OFFLINE: 'Device is offline',
   CXCF_FLAG_CALCERR: 'Formula calculation error',
   CXCF_FLAG_DEFUNCT: 'Defunct channel',
   CXCF_FLAG_OTHEROP: 'Other operator is active',
   CXCF_FLAG_PRGLYCHG: 'Channel was programmatically changed',
   CXCF_FLAG_NOTFOUND: 'Channel not found',
   CXCF_FLAG_COLOR_WEIRD: 'Value is weird',
   CXCF_FLAG_ALARM_ALARM: 'Alarm!',
   CXCF_FLAG_ALARM_RELAX: 'Relaxing after alarm',
   CXCF_FLAG_COLOR_RED: 'Value in red zone',
   CXCF_FLAG_COLOR_YELLOW: 'Value in yellow zone',
}

rflags_color = {
# 'hwerr'
CXRF_CAMAC_NO_X: '#B03060',
CXRF_IO_TIMEOUT: '#B03060',
CXRF_REM_C_PROBL: '#B03060',
CXRF_OVERLOAD: '#B03060',
CXRF_UNSUPPORTED: '#B03060',
# 'sferr'
CXRF_INVAL: '#8B8B00',
CXRF_WRONG_DEV: '#8B8B00',
CXRF_CFG_PROBL: '#8B8B00',
CXRF_DRV_PROBL: '#8B8B00',
CXRF_NO_DRV: '#8B8B00',
CXRF_OFFLINE: '#8B8B00',

CXCF_FLAG_CALCERR: '#8B8B00',
CXCF_FLAG_DEFUNCT: '#4682B4', # 'defunct'
CXCF_FLAG_OTHEROP: '#FFA500', # 'otherop'
CXCF_FLAG_PRGLYCHG: '#D8E3D5', # 'prglychg'
CXCF_FLAG_NOTFOUND: '#404040', # 'notfound'
CXCF_FLAG_COLOR_WEIRD: '#0000FF', # 'weird'
CXCF_FLAG_COLOR_YELLOW: '#EDED6D', # 'yellow'
CXCF_FLAG_COLOR_RED: '#FFC0CB', # 'red'
CXCF_FLAG_ALARM_ALARM: '#FFC0CB', # 'alarm'
CXCF_FLAG_ALARM_RELAX: '#00FF00', # 'alarm_relaxed'

}


cdef class CXRFlagsClass:
    cdef readonly:
        rflags_t CXRF_CAMAC_NO_X
        rflags_t CXRF_CAMAC_NO_Q
        rflags_t CXRF_IO_TIMEOUT
        rflags_t CXRF_REM_C_PROBL
        rflags_t CXRF_OVERLOAD
        rflags_t CXRF_UNSUPPORTED
        rflags_t CXRF_INVAL
        rflags_t CXRF_WRONG_DEV
        rflags_t CXRF_CFG_PROBL
        rflags_t CXRF_DRV_PROBL
        rflags_t CXRF_NO_DRV
        rflags_t CXRF_OFFLINE
        rflags_t CXCF_FLAG_CALCERR
        rflags_t CXCF_FLAG_DEFUNCT
        rflags_t CXCF_FLAG_OTHEROP
        rflags_t CXCF_FLAG_PRGLYCHG
        rflags_t CXCF_FLAG_NOTFOUND
        rflags_t CXCF_FLAG_COLOR_WEIRD
        rflags_t CXCF_FLAG_ALARM_ALARM
        rflags_t CXCF_FLAG_ALARM_RELAX
        rflags_t CXCF_FLAG_COLOR_RED
        rflags_t CXCF_FLAG_COLOR_YELLOW

    def __cinit__(self):
        self.CXRF_CAMAC_NO_X = CXRF_CAMAC_NO_X
        self.CXRF_CAMAC_NO_Q = CXRF_CAMAC_NO_Q
        self.CXRF_IO_TIMEOUT = CXRF_IO_TIMEOUT
        self.CXRF_REM_C_PROBL = CXRF_REM_C_PROBL
        self.CXRF_OVERLOAD = CXRF_OVERLOAD
        self.CXRF_UNSUPPORTED = CXRF_UNSUPPORTED
        self.CXRF_INVAL = CXRF_INVAL
        self.CXRF_WRONG_DEV = CXRF_WRONG_DEV
        self.CXRF_CFG_PROBL = CXRF_CFG_PROBL
        self.CXRF_DRV_PROBL = CXRF_DRV_PROBL
        self.CXRF_NO_DRV = CXRF_NO_DRV
        self.CXRF_OFFLINE = CXRF_OFFLINE
        self.CXCF_FLAG_CALCERR = CXCF_FLAG_CALCERR
        self.CXCF_FLAG_DEFUNCT = CXCF_FLAG_DEFUNCT
        self.CXCF_FLAG_OTHEROP = CXCF_FLAG_OTHEROP
        self.CXCF_FLAG_PRGLYCHG = CXCF_FLAG_PRGLYCHG
        self.CXCF_FLAG_NOTFOUND = CXCF_FLAG_NOTFOUND
        self.CXCF_FLAG_COLOR_WEIRD = CXCF_FLAG_COLOR_WEIRD
        self.CXCF_FLAG_ALARM_ALARM = CXCF_FLAG_ALARM_ALARM
        self.CXCF_FLAG_ALARM_RELAX = CXCF_FLAG_ALARM_RELAX
        self.CXCF_FLAG_COLOR_RED = CXCF_FLAG_COLOR_RED
        self.CXCF_FLAG_COLOR_YELLOW = CXCF_FLAG_COLOR_YELLOW

#serves to bring flags constants to Python if someone need it
CXRFlags = CXRFlagsClass()




from pycda cimport *

import numpy as np
from numpy.polynomial.polynomial import *




class middleChan(object):
    def __init__(self, source_name, middle_name, srclim=None):
        self.source_name = source_name
        self.middle_name = middle_name
        self.srclim = srclim  # limits [min,max]

        self.source_chan = sdchan(source_name)
        self.source_chan.valueChanged.connect(self.source_cb)
        self.middle_chan = sdchan(middle_name)
        self.middle_chan.valueChanged.connect(self.middle_cb)

        self.src_changing = False
        self.mid_changing = False

        self.starting = True

    def source_cb(self, chan, value):
        if self.starting:
            self.starting = False
        if self.mid_changing:
            self.mid_changing = False
            return
        self.middle_chan.setValue(self.src2mid(value))
        self.src_changing = True

    def middle_cb(self, chan, value):
        if self.starting:
            self.starting = False
            return
        if self.src_changing:
            self.src_changing = False
            return
        srcval = self.mid2src(value)
        if self.srclim is not None:
            if srcval < self.srclim[0] and self.srclim[0] is not None:
                self.middle_chan.setValue(self.src2mid(self.srclim[0]))
                self.source_chan.setValue(self.srclim[0])
                return
            if srcval > self.srclim[1] and self.srclim[1] is not None:
                self.middle_chan.setValue(self.src2mid(self.srclim[1]))
                self.source_chan.setValue(self.srclim[1])
                return
        self.source_chan.setValue(srcval)
        self.mid_changing = True

    def src2mid(self, src_val):
        return src_val

    def mid2src(self, mid_val):
        return mid_val


class middleChanPoly(middleChan):
    def __init__(self, source_name, middle_name, srclim, s2m, m2s):
        super(middleChanPoly, self).__init__(source_name, middle_name, srclim)
        self.s2m = s2m
        self.m2s = m2s

    def src2mid(self, src_val):
        mid_val = polyval(src_val, self.s2m)
        return mid_val

    def mid2src(self, mid_val):
        src_val = polyval(mid_val, self.m2s)
        return src_val


class middleChanRO(object):
    def __init__(self, source_name, middle_name):
        self.source_name = source_name
        self.middle_name = middle_name

        self.source_chan = cxchan(source_name)
        self.source_chan.valueChanged.connect(self.source_cb)
        self.middle_chan = cxchan(middle_name)

    def source_cb(self, chan, value):
        self.middle_chan.setValue(self.src2mid(value))

    def src2mid(self, src_val):
        return src_val


class middleChanPolyRO(middleChanRO):
    def __init__(self, source_name, middle_name, s2m):
        super(middleChanPolyRO, self).__init__(source_name, middle_name)
        self.s2m = s2m  #i2f coefs array

    def src2mid(self, src_val):
        mid_val = polyval(src_val, self.s2m)
        return mid_val


class middleChanEpics:
    def __init__(self, source_name, middle_name):
        self.source_name = source_name
        self.middle_name = middle_name
        self.src_changing = False
        self.mid_changing = False
        self.starting = False

        self.source_chan = camonitor(source_name, self.source_cb)
        self.middle_chan = cxchan(middle_name)
        self.middle_chan.valueChanged.connect(self.middle_cb)

    def source_cb(self, value):
        if self.starting:
            self.starting = False
        if self.mid_changing:
            self.mid_changing = False
            return
        self.middle_chan.setValue(self.src2mid(value))
        self.src_changing = True

    def middle_cb(self, chan, value):
        if self.starting:
            return
        if self.src_changing:
            self.src_changing = False
            return
        caput(self.source_name, self.mid2src(value))
        self.mid_changing = True

    def src2mid(self, src_val):
        return src_val

    def mid2src(self, mid_val):
        return mid_val



# Cython wrapper for cxv4 cda
# by Fedor Emanov

from .misc_types cimport *
from .cx_common_types cimport *
from .cx cimport *


cdef extern from "cda.h":
    enum:
        CDA_PATH_MAX

    ctypedef CxDataRef_t cda_dataref_t
    enum:
        CDA_DATAREF_ERROR
        CDA_DATAREF_NONE

    ctypedef int cda_context_t
    enum:
        CDA_CONTEXT_ERROR

    ctypedef int cda_varparm_t
    enum:
        CDA_VARPARM_ERROR

    enum:
        CDA_CONTEXT_OPT_NONE
        CDA_CONTEXT_OPT_NO_OTHEROP
        CDA_CONTEXT_OPT_IGN_UPDATE

    enum:
        CDA_DATAREF_OPT_NONE
        CDA_DATAREF_OPT_PRIVATE
        CDA_DATAREF_OPT_NO_RD_CONV
        CDA_DATAREF_OPT_SHY
        CDA_DATAREF_OPT_FIND_ONLY
        CDA_DATAREF_OPT_ON_UPDATE
        CDA_DATAREF_OPT_rsrvd26
        CDA_DATAREF_OPT_NO_WR_WAIT
        CDA_DATAREF_OPT_rsrvd24
        CDA_DATAREF_OPT_DEBUG
        CDA_DATAREF_OPT_PRIV_SRV
        CDA_DATAREF_OPT_EXCLUSIVE
        CDA_DATAREF_OPT_NOMONITOR

    enum:
        CDA_OPT_NONE
        CDA_OPT_IS_W
        CDA_OPT_READONLY
        CDA_OPT_HAS_PARAM
        CDA_OPT_RETVAL_RQD
        CDA_OPT_DO_EXEC
        CDA_OPT_RD_FLA
        CDA_OPT_WR_FLA

    enum:
        CDA_PROCESS_SEVERE_ERR
        CDA_PROCESS_ERR
        CDA_PROCESS_DONE
        CDA_PROCESS_FLAG_BUSY
        CDA_PROCESS_FLAG_REFRESH

    # context events
    enum:
        CDA_CTX_R_CYCLE
        CDA_CTX_EVMASK_CYCLE
        CDA_CTX_R_SRVSTAT
        CDA_CTX_EVMASK_SRVSTAT
        CDA_CTX_R_NEWSRV
        CDA_CTX_EVMASK_NEWSRV

    # ref events
    enum:
        CDA_REF_R_UPDATE
        CDA_REF_R_STATCHG
        CDA_REF_R_STRSCHG
        CDA_REF_R_RDSCHG
        CDA_REF_R_FRESHCHG
        CDA_REF_R_QUANTCHG
        CDA_REF_R_RANGECHG
        CDA_REF_R_RSLVSTAT
        CDA_REF_R_CURVAL
        CDA_REF_R_LOCKSTAT

        CDA_REF_EVMASK_UPDATE
        CDA_REF_EVMASK_STATCHG
        CDA_REF_EVMASK_STRSCHG
        CDA_REF_EVMASK_RDSCHG
        CDA_REF_EVMASK_FRESHCHG
        CDA_REF_EVMASK_QUANTCHG
        CDA_REF_EVMASK_RANGECHG
        CDA_REF_EVMASK_RSLVSTAT
        CDA_REF_EVMASK_CURVAL
        CDA_REF_EVMASK_LOCKSTAT

    enum:
        CDA_LOCK_RLS
        CDA_LOCK_SET

    ctypedef enum cda_serverstatus_t:
        #/* Note: the statuses are ordered by decreasing of severity,
        #  so that a "most severe of several" status can be easily chosen */
        CDA_SERVERSTATUS_ERROR
        CDA_SERVERSTATUS_DISCONNECTED
        CDA_SERVERSTATUS_FROZEN
        CDA_SERVERSTATUS_ALMOSTREADY
        CDA_SERVERSTATUS_NORMAL

    enum:
        CDA_RSLVSTAT_NOTFOUND
        CDA_RSLVSTAT_SEARCHING
        CDA_RSLVSTAT_FOUND

    int cda_ref_is_sensible(cda_dataref_t ref)

    ctypedef void (*cda_context_evproc_t)(int uniq, void *privptr1, cda_context_t cid, int reason, int info_int, void *privptr2)

    ctypedef void (*cda_dataref_evproc_t)(int uniq, void *privptr1, cda_dataref_t ref, int reason, void *info_ptr, void *privptr2)

    # context management
    cda_context_t cda_new_context(int uniq,   void *privptr1,
                               const char           *defpfx, int   flags,
                               const char           *argv0,
                               int                   evmask,
                               cda_context_evproc_t  evproc,
                               void                 *privptr2)
    int cda_del_context(cda_context_t cid)

    int cda_add_context_evproc(cda_context_t cid,
                               int evmask,
                               cda_context_evproc_t evproc,
                               void *privptr2)

    int cda_del_context_evproc(cda_context_t cid,
                               int evmask,
                               cda_context_evproc_t evproc,
                               void *privptr2)

    # Channels management
    cda_dataref_t  cda_add_chan(cda_context_t         cid,
                               const char           *base,
                               const char           *spec,
                               int                   flags,
                               cxdtype_t             dtype,
                               int                   max_nelems,
                               int                   evmask,
                               cda_dataref_evproc_t  evproc,
                               void                 *privptr2)

    int            cda_del_chan(cda_dataref_t ref)
    int            cda_set_type(cda_dataref_t ref, cxdtype_t dtype, int max_nelems)

    int            cda_add_dataref_evproc(cda_dataref_t     ref,
                                      int                   evmask,
                                      cda_dataref_evproc_t  evproc,
                                      void                 *privptr2)

    int            cda_del_dataref_evproc(cda_dataref_t     ref,
                                      int                   evmask,
                                      cda_dataref_evproc_t  evproc,
                                      void                 *privptr2)

    int cda_lock_chans(int count, cda_dataref_t *refs, int operation, int lockset_id)

    char *cda_combine_base_and_spec(cda_context_t     cid,
                                const char           *base,
                                const char           *spec,
                                char                 *namebuf,
                                size_t                namebuf_size)

    # Simplified channels API

    cda_dataref_t cda_add_dchan(cda_context_t cid, const char *name)
    int cda_get_dcval(cda_dataref_t ref, double *v_p)
    int cda_set_dcval(cda_dataref_t ref, double  val)

    cda_dataref_t cda_add_ichan(cda_context_t cid, const char *name)
    int cda_get_icval(cda_dataref_t ref, int *v_p)
    int cda_set_icval(cda_dataref_t ref, int val)

    int cda_add_schan(cda_context_t cid, const char *name, size_t maxlen)
    int cda_get_scval(cda_dataref_t ref, char *buf, size_t bufsize, size_t *len_p)
    int cda_set_scval(cda_dataref_t ref, char *str, size_t len)

    # Formulae management
    cda_dataref_t cda_add_formula(cda_context_t         cid,
                                  const char           *base,
                                  const char           *spec,
                                  int                   options,
                                  CxKnobParam_t        *params,
                                  int                   num_params,
                                  int                   evmask,
                                  cda_dataref_evproc_t  evproc,
                                  void                 *privptr2)

    cda_dataref_t cda_add_varchan(cda_context_t cid, const char *varname)
    cda_varparm_t cda_add_varparm(cda_context_t cid, const char *varname)

    #/**/
    int cda_src_of_ref           (cda_dataref_t ref, const char **src_p)
    int cda_dtype_of_ref         (cda_dataref_t ref)
    int cda_max_nelems_of_ref    (cda_dataref_t ref)
    int cda_current_nelems_of_ref(cda_dataref_t ref)
    int cda_fresh_age_of_ref     (cda_dataref_t ref, cx_time_t *fresh_age_p)
    int cda_quant_of_ref         (cda_dataref_t ref, CxAnyVal_t *q_p, cxdtype_t *q_dtype_p)
    int cda_range_of_ref         (cda_dataref_t ref, CxAnyVal_t range[], cxdtype_t *range_dtype_p);
    int cda_current_dtype_of_ref (cda_dataref_t ref)

    int cda_strings_of_ref       (cda_dataref_t  ref,
                                  char    **ident_p,
                                  char    **label_p,
                                  char    **tip_p,
                                  char    **comment_p,
                                  char    **geoinfo_p,
                                  char    **rsrvd6_p,
                                  char    **units_p,
                                  char    **dpyfmt_p)
    int cda_phys_rds_of_ref      (cda_dataref_t  ref,
                                  int     *phys_count_p,
                                  double  **rds_p)

    int cda_hwinfo_of_ref        (cda_dataref_t  ref,
                                  int       *rw_p,
                                  cxdtype_t *dtype_p,
                                  int       *nelems_p,
                                  int       *srv_hwid_p,
                                  int       *cln_hwr_p)

    int cda_lock_stat_of_ref(cda_dataref_t ref)
    int cda_status_of_ref_sid(cda_dataref_t ref)

    int cda_status_srvs_count(cda_context_t  cid)
    cda_serverstatus_t cda_status_of_srv(cda_context_t cid, int nth)
    const char *cda_status_srv_scheme(cda_context_t  cid, int nth)
    const char *cda_status_srv_name(cda_context_t cid, int nth)

    int cda_srvs_of_ref(cda_dataref_t ref, uint8 *conns_u, int conns_u_size)
    int cda_reconnect_srv(cda_context_t cid, int nth)
    int cda_add_server_conn(cda_context_t cid, const char *srvref)

    #/**********************/
    int cda_process_ref(cda_dataref_t ref, int options,
                        double userval,
                        CxKnobParam_t *params, int num_params)
    int cda_get_ref_dval(cda_dataref_t ref,
                         double     *curv_p,
                         CxAnyVal_t *curraw_p, cxdtype_t *curraw_dtype_p,
                         rflags_t *rflags_p, cx_time_t *timestamp_p)
    int cda_get_ref_ival(cda_dataref_t ref,
                         int        *curv_p,
                         CxAnyVal_t *curraw_p, cxdtype_t *curraw_dtype_p,
                         rflags_t   *rflags_p, cx_time_t *timestamp_p)


    int cda_rd_convert  (cda_dataref_t ref, double raw, double *result_p)
    int cda_r_convert   (cda_dataref_t ref, double raw, double *result_p)
    int cda_snd_ref_data(cda_dataref_t ref, cxdtype_t dtype, int nelems, void *data)
    int cda_get_ref_data(cda_dataref_t ref, size_t ofs, size_t size, void *buf)
    int cda_get_ref_stat(cda_dataref_t ref, rflags_t *rflags_p, cx_time_t *timestamp_p)
    int cda_acc_ref_data(cda_dataref_t ref, void **buf_p, size_t *size_p)

    int cda_stop_formula(cda_dataref_t ref)

    void cda_allow_plugins_loading(int allow)


    const char *cda_strserverstatus_short(cda_serverstatus_t status)
    char *cda_last_err()

    void cda_do_cleanup(int uniq)



# data types declarations and utility functions

DTYPE_UNKNOWN = CXDTYPE_UNKNOWN
DTYPE_INT8 = CXDTYPE_INT8
DTYPE_INT16 = CXDTYPE_INT16
DTYPE_INT32 = CXDTYPE_INT32
DTYPE_INT64 = CXDTYPE_INT64
DTYPE_UINT8 = CXDTYPE_UINT8
DTYPE_UINT16 = CXDTYPE_UINT16
DTYPE_UINT32 = CXDTYPE_UINT32
DTYPE_UINT64 = CXDTYPE_UINT64
DTYPE_SINGLE = CXDTYPE_SINGLE
DTYPE_DOUBLE = CXDTYPE_DOUBLE
DTYPE_TEXT = CXDTYPE_TEXT
DTYPE_UCTEXT = CXDTYPE_UCTEXT

cx_dtype_map = {
    'int8': CXDTYPE_INT8,
    'int16': CXDTYPE_INT16,
    'int32': CXDTYPE_INT32,
    'int': CXDTYPE_INT32,
    'int64': CXDTYPE_INT64,
    'uint8': CXDTYPE_UINT8,
    'uint16': CXDTYPE_UINT16,
    'uint32': CXDTYPE_UINT32,
    'uint': CXDTYPE_UINT32,
    'uint64': CXDTYPE_UINT64,
    'single': CXDTYPE_SINGLE,
    'float': CXDTYPE_SINGLE,
    'double': CXDTYPE_DOUBLE,
    'text': CXDTYPE_TEXT,
    'str': CXDTYPE_TEXT,
    'uctext': CXDTYPE_UCTEXT,
    #'unknown': CXDTYPE_UNKNOWN,
}

ctypedef fused atype:
    float64
    float32
    int32
    int16
    int8
    uint32
    uint16
    uint8
    char8
    char32
    #int64 yet unsupported
    #uint64  yet unsupported
# currently int64 generates the same func prototype as int and will not work
# that's cython behavior, not cpp for example

# this needed for independent typing of 2-args functions
ctypedef fused atype2:
    float64
    float32
    int32
    int16
    int8
    uint32
    uint16
    uint8
    char8
    char32
    #int64 yet unsupported
    #uint64  yet unsupported



# function for pythonize cx-any-val
# not to effective, but not a bottleneck
cdef:
    aval_value(CxAnyVal_t *aval, cxdtype_t dtype):
        if dtype == CXDTYPE_DOUBLE: return aval.f64
        elif dtype == CXDTYPE_INT32:  return aval.i32
        elif dtype == CXDTYPE_INT8:   return aval.i8
        elif dtype == CXDTYPE_INT16:  return aval.i16
        elif dtype == CXDTYPE_INT64:  return aval.i64
        elif dtype == CXDTYPE_UINT8:  return aval.u8
        elif dtype == CXDTYPE_UINT16: return aval.u16
        elif dtype == CXDTYPE_UINT32: return aval.u32
        elif dtype == CXDTYPE_UINT64: return aval.u64
        elif dtype == CXDTYPE_SINGLE: return aval.f32
        elif dtype == CXDTYPE_TEXT:   return aval.c8
        elif dtype == CXDTYPE_UCTEXT: return aval.c32
        return None

    void aval_setvalue(value, CxAnyVal_t *aval, cxdtype_t dtype):
        if dtype == CXDTYPE_DOUBLE: aval.f64 = value
        elif dtype == CXDTYPE_INT32:  aval.i32 = value
        elif dtype == CXDTYPE_INT8:   aval.i8 = value
        elif dtype == CXDTYPE_INT16:  aval.i16 = value
        elif dtype == CXDTYPE_INT64:  aval.i64 = value
        elif dtype == CXDTYPE_UINT8:  aval.u8 = value
        elif dtype == CXDTYPE_UINT16: aval.u16 = value
        elif dtype == CXDTYPE_UINT32: aval.u32 = value
        elif dtype == CXDTYPE_UINT64: aval.u64 = value
        elif dtype == CXDTYPE_SINGLE: aval.f32 = value
        elif dtype == CXDTYPE_TEXT:   aval.c8 = value
        elif dtype == CXDTYPE_UCTEXT: aval.c32 = value

    void aval_set(atype value,  CxAnyVal_t *aval):
        if atype is float64: aval.f64 = value
        if atype is float32: aval.f32 = value
        if atype is int64: aval.i64 = value
        if atype is int32: aval.i32 = value
        if atype is int16: aval.i16 = value
        if atype is int8: aval.i8 = value
        if atype is uint64: aval.u64 = value
        if atype is uint32: aval.u32 = value
        if atype is uint16: aval.u16 = value
        if atype is uint8: aval.u8 = value
        if atype is char8: aval.c8 = value
        if atype is char32: aval.c32 = value

    str cxdtype2pycode(cxdtype_t dtype):
        if dtype == CXDTYPE_DOUBLE: return 'd'
        elif dtype == CXDTYPE_SINGLE: return 'f'
        elif dtype == CXDTYPE_INT32:  return 'l'
        elif dtype == CXDTYPE_INT8:   return 'b'
        elif dtype == CXDTYPE_INT16:  return 'h'
        elif dtype == CXDTYPE_INT64:  return 'q'
        elif dtype == CXDTYPE_UINT8:  return 'B'
        elif dtype == CXDTYPE_UINT16: return 'H'
        elif dtype == CXDTYPE_UINT32: return 'L'
        elif dtype == CXDTYPE_UINT64: return 'Q'
        return None





# general channel (any cx type)
cdef class Chan(BaseChan):
    cdef:
        readonly object val, prev_val

    # def __cinit__(self, str name, **kwargs):
    #     #BaseChan.__init__(self, name, **kwargs)
    #     pass

    cdef void cb(self):
        cdef CxAnyVal_t aval
        self.prev_val = self.val
        self.get_data(0, self.itemsize, <void*>&aval)
        self.val = aval_value(&aval, self.dtype)
        if self.val != self.prev_val or self.first_cycle:
            self.valueChanged.emit(self)
            self.first_cycle = False
        self.valueMeasured.emit(self)

    cpdef void setValue(self, value):
        cdef CxAnyVal_t aval
        aval_setvalue(value, &aval, self.dtype)
        self.snd_data(self.dtype, 1, <void*>&aval)



from cx4.cxscheduler cimport sl_main_loop, sl_break
cimport cx4.cxscheduler as sl

SL_RD = sl.SL_RD
SL_WR = sl.SL_WR
SL_EX = sl.SL_EX
SL_CE = sl.SL_CE


cpdef main_loop():
    sl_main_loop()

cpdef break_():
    sl_break()





cimport cython
from libc.stdlib cimport realloc, free, malloc
from libc.string cimport memmove
from cpython cimport array

cdef bint using_numpy
try:
    import numpy as np
    using_numpy = True
except ImportError:
    using_numpy = False

from cx4.cx cimport *
from cx4.cda cimport *


cdef class RslvStatsClass:
    cdef readonly:
        long found
        long searching
        long notfound
    def __cinit__(self):
        self.found = CDA_RSLVSTAT_FOUND
        self.searching = CDA_RSLVSTAT_SEARCHING
        self.notfound = CDA_RSLVSTAT_NOTFOUND

RslvStats = RslvStatsClass()




from posix.types cimport time_t, suseconds_t

# Dew to somehow incorrect timeval definition in posix.time
# if timeval in posix.time - not a type
ctypedef struct timeval:
        time_t      tv_sec
        suseconds_t tv_usec


cdef extern from "cxscheduler.h" nogil:
    enum:
       SL_RD  #,  // Watch when descriptor is ready for read
       SL_WR  #,  // Watch when descriptor is ready for write
       SL_EX  #,  // Watch descriptor for exceptions (in fact -- OOB data)
       SL_CE  #,  // Watch for Connect Errors.  That's a hint for Xh_cxscheduler exclusively

    ctypedef int sl_tid_t
    ctypedef int sl_fdh_t

    ctypedef void (*sl_tout_proc)(int uniq, void *privptr1, sl_tid_t tid, void *privptr2)
    ctypedef void (*sl_fd_proc)(int uniq, void *privptr1, sl_fdh_t fdh, int fd, int mask, void *privptr2)

    sl_tid_t sl_enq_tout_at(int uniq, void *privptr1,
                                 timeval *when,
                                 sl_tout_proc cb, void *privptr2)

    sl_tid_t sl_enq_tout_after(int uniq, void *privptr1,
                                int             usecs,
                                sl_tout_proc cb, void *privptr2)

    int sl_deq_tout(sl_tid_t tid)

    sl_fdh_t sl_add_fd(int uniq, void *privptr1, int fd, int mask, sl_fd_proc cb, void *privptr2)
    int sl_del_fd(sl_fdh_t fdh)
    int sl_set_fd_mask(sl_fdh_t fdh, int mask)

    int sl_main_loop()
    int sl_break()

    #/* Note:
    #   the sl_set_select_behaviour() is implemented in
    #   CLASSIC, select-based cxscheduler.c only */

    ctypedef void (*sl_at_select_proc)()

    int sl_set_select_behaviour(sl_at_select_proc  before,
                             sl_at_select_proc  after,
                             int usecs_idle)

    ctypedef void (*sl_on_timeback_proc)()
    int sl_set_on_timeback_proc(sl_on_timeback_proc proc)

    ctypedef int (*sl_uniq_checker_t)(const char *func_name, int uniq)

    void sl_set_uniq_checker(sl_uniq_checker_t checker)
    void sl_do_cleanup(int uniq)



import sys
import importlib

__all__ = ['QT_LIB']
DChan = None

qts = ['PyQt5', 'PyQt4']
QT_LIB = None
cda = None

for lib in qts:
    if lib in sys.modules:
        QT_LIB = lib
        break
if QT_LIB is None:
    for lib in qts:
        try:
            importlib.import_module(lib)
            QT_LIB = lib
            break
        except ImportError:
            pass
if QT_LIB is None:
    ImportError("PyQt not found. it's required for qcda")

if QT_LIB == 'PyQt5':
    import pycx4.q5cda
    __all__ += pycx4.q5cda.__all__
    # get q5cda to qcda namespace
    from pycx4.q5cda import *
elif QT_LIB == 'PyQt4':
    import pycx4.q4cda
    __all__ += pycx4.q4cda.__all__
    # get q5cda to qcda namespace
    from pycx4.q4cda import *
else:
    ImportError("something wrong")





# C callback for context
cdef void evproc_cont_cycle(int uniq, void *privptr1, cda_context_t cid, int reason,
                      int info_int, void *privptr2) with gil:
    cdef Context cont=<Context>(<event*>privptr2).objptr
    cont.serverCycle.emit(cont)


cdef class Context(CdaObject):
    cdef readonly:
        cda_context_t cid
        str defpfx
    cdef public:
        object serverCycle
    cdef:
        void **chans
        int channum

    def __cinit__(self, defpfx="cx::", **kwargs):
        cdef:
            int ret
            int options = 0
            char *c_defpfx
        ascii_pfx = defpfx.encode("ascii") # encode to ascii
        c_defpfx = ascii_pfx  # convert to char*

        if 'other_opp_flag' in kwargs:
            if kwargs['other_opp_flag']:
                options += CDA_CONTEXT_OPT_NO_OTHEROP

        if 'ignore_update' in kwargs:
            if kwargs['ignore_update']:
                options += CDA_CONTEXT_OPT_IGN_UPDATE

        ret = cda_new_context(0, NULL, c_defpfx, options, NULL, 0, <cda_context_evproc_t>NULL, NULL)
        self.check_exception(ret)
        self.cid, self.defpfx, self.chans, self.channum = ret, defpfx, NULL, 0

        self.serverCycle = Signal(object)

    def __dealloc__(self):
        if self.cid > 0:
            self.check_exception( cda_del_context(self.cid) )
            self.cid = 0
        if self.channum > 0:
            free(self.chans)
            self.chans = NULL
            self.channum = 0

    def __str__(self):
        return '<CdaContext: cid=%d, defpfx=%s, channum=%d>' % (self.cid, self.defpfx, self.channum)

    def enable_serverCycle(self):
        self.add_event(CDA_CTX_EVMASK_CYCLE, <void*>evproc_cont_cycle, <void*>self, NULL)

    cdef void save_chan(self, void *chan):
        cdef:
            void *tmp = realloc(<void*>self.chans, sizeof(void*) * (self.channum+1))
        if not tmp: raise MemoryError()
        self.chans = <void**>tmp
        self.chans[self.channum] = chan
        self.channum += 1

    cdef void drop_chan(self, void *chan):
        if not self.cid: return
        cdef:
            void *tmp
            int ind
        for ind in range(self.channum):
            if chan == self.chans[ind]:
                if self.channum == 1:
                    free(self.chans)
                    self.chans = NULL
                else:
                    memmove(&(self.chans[ind]), &(self.chans[self.channum-1]), sizeof(void*) )
                    tmp = realloc(<void*>self.chans, sizeof(void*) * (self.channum-1))
                    if not tmp and self.channum>1: raise MemoryError()
                    self.chans = <void**>tmp
                self.channum -= 1
                return

    cdef void register_event(self, event *ev):
        self.check_exception( cda_add_context_evproc(self.cid, ev.evmask, <cda_context_evproc_t>ev.evproc, ev) )

    cdef void unregister_event(self, event *ev):
        self.check_exception( cda_del_context_evproc(self.cid, ev.evmask, <cda_context_evproc_t>ev.evproc, ev) )

    cdef void check_exception(self, int c_res):
        if c_res < 0:
            raise Exception("cda context error: cid=%s, %s, errcode=%s" % (self.cid, cda_last_err(), c_res ))




cdef class ChanFactory:
    cdef:
        dict single_chans, vector_chans

    def __init__(self):
        self.single_chans = {
            'double': DChan,
            'int': IChan,
            'int32': IChan
        }
        self.vector_chans = {
            'str': StrChan,
        }

    def create_chan(self, name, dtype, dsize, **kwargs):
        if dtype not in cx_dtype_map:
            return None
        if dsize == 1:
            if dtype in self.single_chans:
                return self.single_chans[dtype](name, **kwargs)
            else:
                return Chan(name, cx_dtype_map[dtype])
        elif dtype in self.vector_chans:
            return self.vector_chans[dtype](name, dsize)
        return VChan(name, cx_dtype_map[dtype], dsize, **kwargs)

cfactory = ChanFactory()




import os
import pwd

def cx_installpath():
    search_pathes = []
    print(os.getenv('HOME'))

    if os.getenv('CXDIR'):
        search_pathes.append(os.getenv('CXDIR') + '/4cx/exports')
        search_pathes.append(os.getenv('CXDIR') + '/exports')

    search_pathes.append(os.getenv('HOME') + '/cx/4cx/exports')

    if os.getenv('USER') == "root":
        search_pathes.append(pwd.getpwnam(os.getenv('USER')).pw_dir + '/4pult')
        search_pathes.append(pwd.getpwnam(os.getenv('USER')).pw_dir + '/cx/4cx/exports')

    try:
        search_pathes.append(pwd.getpwnam('oper').pw_dir + '/4pult')
    except KeyError:
        pass

    try:
        search_pathes.append(pwd.getpwnam('pi').pw_dir + '/4pult')
    except KeyError:
        pass

    # if this subdirs are exist - let's say it' valid cx install place
    subdirs = ['/include', '/lib']
    for basedir in search_pathes:
        count = 0
        for subdir in subdirs:
            if os.path.exists(basedir + subdir):
                count += 1
        if count == len(subdirs):
            return basedir
    return None


if __name__ == '__main__':

    print('using cx install path:', cx_installpath())





# vector-data channel class
cdef class VChan(BaseChan):
    cdef:
        readonly object val # showing data to numpy
        readonly array.array aval
        readonly int nelems

    def __cinit__(self, str name, **kwargs):
        self.aval = array.array(cxdtype2pycode(self.dtype))
        self.resize_val(1)

    cdef void resize_val(self, int size):
        array.resize(self.aval, size)
        if using_numpy:
            self.val = np.asarray(self.aval)
        self.nelems = size

    cdef void cb(self):
        cdef int c_len
        cdef int nelems = self.current_nelems()
        if self.nelems != nelems:
            self.resize_val(nelems)
        c_len = self.get_data(0, self.itemsize * nelems, self.aval.data.as_voidptr)
        #nelems_read = int(c_len/self.itemsize)
        self.valueMeasured.emit(self)
        self.valueChanged.emit(self)

    cpdef void setValue(self, atype[::1] value):
        if value.size > self.max_nelems:
            raise Exception('value size greater than channel.max_nelems')
        cdef cxdtype_t dtype
        if atype is float64: dtype = CXDTYPE_DOUBLE
        elif atype is float32: dtype = CXDTYPE_SINGLE
        elif atype is int32: dtype = CXDTYPE_INT32
        elif atype is int16: dtype = CXDTYPE_INT16
        elif atype is int8: dtype = CXDTYPE_INT8
        elif atype is uint32: dtype = CXDTYPE_UINT32
        elif atype is uint16: dtype = CXDTYPE_UINT16
        elif atype is uint8: dtype = CXDTYPE_UINT8
        elif atype is char8: dtype = CXDTYPE_TEXT
        elif atype is char32: dtype = CXDTYPE_UCTEXT
        self.snd_data(dtype, value.size, <void*>&value[0])

    # older version can work with int64, not yet totally removed
    # cpdef void setValue(self, np.ndarray value):
    #     if value.size > self.max_nelems:
    #         raise Exception('value size greater than channel.max_nelems')
    #     dtype = np2cxdtype(value.dtype)
    #     self.snd_data(dtype, value.size, <void*>value.data)



__all__ = ['Context', 'StrChan', 'DChan', 'IChan', 'Chan', 'VChan', 'VPChan',
           'Timer', 'Signal',
           'DTYPE_UNKNOWN', 'DTYPE_INT8', 'DTYPE_INT16', 'DTYPE_INT32', 'DTYPE_INT64', 'DTYPE_UINT8','DTYPE_UINT16',
           'DTYPE_UINT32', 'DTYPE_UINT64', 'DTYPE_SINGLE', 'DTYPE_DOUBLE', 'DTYPE_TEXT', 'DTYPE_UCTEXT',
           'ChanFactory', 'cfactory',
           'rflags_order', 'rflags_meaning', 'rflags_color_status',
           'CXRFlags', 'RslvStats'
    ]


include 'imports.pxi'
include 'scheduler.pxi'
include 'signal.pxi'
include 'timer.pxi'
include 'fd_event.pxi'
include 'cda_common.pxi'
include 'gw.pxi'


__all__ += ['main_loop', 'break_', 'FdEvent', 'PassGW', "SL_RD", "SL_WR", "SL_EX", "SL_CE"]






# cx-cx gateway and basic processing channels

# pass value
cdef class PassGW:
    cdef:
        object in_chan, out_chan, last_val

    def __init__(self, str in_name, str out_name, **kwargs):
        dtype = kwargs.get('dtype', 'double')
        max_nelems = kwargs.get('max_nelems', 1)

        self.in_chan = cfactory.create_chan(in_name, dtype, max_nelems, **kwargs)
        self.out_chan = cfactory.create_chan(out_name, dtype, max_nelems, **kwargs)

        self.in_chan.valueMeasured.connect(self.in_c_proc)
        self.out_chan.valueMeasured.connect(self.out_c_proc)

        self.last_val = None

    cpdef void in_c_proc(self, chan):
        self.last_val = self.in_chan.val
        self.out_chan.setValue(self.in_chan.val)

    cpdef void out_c_proc(self, chan):
        if chan.val == self.last_val or self.last_val is None:
            return
        self.out_chan.setValue(self.last_val)


cdef class ProcGW:
    cdef:
        object in_chan, out_chan

    def __init__(self, str in_name, str out_name, **kwargs):
        dtype = kwargs.get('dtype', 'double')
        max_nelems = kwargs.get('max_nelems', 1)

        self.in_chan = cfactory.create_chan(in_name, dtype, max_nelems, **kwargs)
        self.out_chan = cfactory.create_chan(out_name, dtype, max_nelems, **kwargs)

        self.in_chan.valueMeasured.connect(self.in_c_proc)
        self.out_chan.valueMeasured.connect(self.out_c_proc)

    cpdef void in_c_proc(self, chan):
        pass

    cpdef void out_c_proc(self, chan):
        pass

    def in2out(in_val):
        pass



# Signal container to bypass inheritance from QObject
# to be encapsulated in Qt channel-like classes

class SigContainer(QObject):
    def __new__(cls, *args):
        sig = pyqtSignal(args)
        cont = type('sig_cont', (QObject, ), {
                 'sig': pyqtSignal(args)})
        inst = cont()
        inst.emit, inst.connect, inst.disconnect = inst.sig.emit, inst.sig.connect, inst.sig.disconnect
        return inst


Signal = SigContainer






# Utility functions to avoid dependencies
# this functions do not perform any checks... to be used with precautions
# # v * a inplace
# def vector_mult(atype[::1] v, double a):
#     cdef int i
#     for i in range(len(v)):
#         v[i] = <atype>(v[i] * a)
#
# def vector_iadd(atype[::1] v1, atype2[::1] v2):
#     cdef int i
#     for i in range(len(v1)):
#         v1[i] += <atype>v2[i]
#
# def vector_isub(double[::1] v1, double[::1] v2):
#     cdef int i
#     for i in range(len(v1)):
#         v1[i] -= v2[i]
#
# def vector_sub(double[::1] vout, atype[::1] v1, double[::1] v2):
#     cdef int i
#     for i in range(len(v1)):
#         vout[i] = <double>v1[i] - v2[i]
#
# def avg(atype[::1] vin, int wi, int wf):
#     cdef:
#         int i
#         double a=0
#     for i in range(wi, wf):
#         a += vin[i]
#     a /= wf-wi
#     return a

# this supports just double, but faster and lighter
# v * a inplace
cdef vector_mult(double[::1] v, double a):
    cdef int i
    for i in range(len(v)):
        v[i] = (v[i] * a)

cdef vector_iadd(double[::1] v1, double[::1] v2):
    cdef int i
    for i in range(len(v1)):
        v1[i] += v2[i]

cdef vector_isub(double[::1] v1, double[::1] v2):
    cdef int i
    for i in range(len(v1)):
        v1[i] -= v2[i]

cdef vector_sub(double[::1] vout, double[::1] v1, double[::1] v2):
    cdef int i
    for i in range(len(v1)):
        vout[i] = v1[i] - v2[i]

cdef avg(double[::1] vin, int wi, int wf):
    cdef:
        int i
        double a=0
    for i in range(wi, wf):
        a += vin[i]
    a /= wf-wi
    return a



# vector-data channel class with some data processing capabilities
cdef class VPChan(VChan):
    cdef:
        readonly object avg_val, bg_val
        readonly array.array avg_aval, bg_aval, bg
        readonly object avgReady, bgReady # extra signals
        readonly int reg_s, reg_f
        readonly double reg_avg

        int n_avg, avg_count
        int bg_count, bg_anum
        bint change_sign, dont_update, bg_apply, bg_ready

    def __cinit__(self, str name, **kwargs):
        self.change_sign = kwargs.get('change_sign', False)
        self.avgReady = Signal(object)
        self.bgReady = Signal(object)
        self.bg_ready = False
        self.bg_count = -1
        self.n_avg = 1
        self.avg_aval = array.array('d')

    cdef void resize_avg_val(self, int size):
        array.resize(self.avg_aval, size)
        if using_numpy:
            self.avg_val = np.asarray(self.avg_aval)
        self.avg_count = 0 # this to reset averaging

    cdef void reset_bg(self):
        self.bg_ready = False
        self.bg = None

    cdef void cb(self):
        cdef int nelems = self.current_nelems()
        if self.nelems != nelems:
            self.resize_val(nelems)
            if self.n_avg > 1:
                self.resize_avg_val(nelems)
            if self.bg_ready:
                self.reset_bg()
        c_len = self.get_data(0, self.itemsize * nelems, self.aval.data.as_voidptr)

        if self.change_sign:
            vector_mult(self.aval, -1.0)

        if self.bg_count > 0:
            vector_iadd(self.bg, self.aval)
            self.bg_count -= 1
            print("getting bg, steps left: ", self.bg_count)
        if self.bg_count == 0 and not self.bg_ready:
            vector_mult(self.bg, 1.0/self.bg_anum)
            self.bg_ready = True
            self.bgReady.emit(self)

        if self.bg_apply and self.bg_ready:
            vector_sub(self.bg_aval, self.aval, self.bg)

        if self.n_avg > 1:
            if self.avg_count == 0:
                array.zero(self.avg_aval)
            vector_iadd(self.avg_aval, self.aval)
            self.avg_count += 1
            if self.avg_count == self.n_avg:
                vector_mult(self.avg_aval, 1.0/self.n_avg)
                if self.bg_apply:
                    vector_isub(self.avg_aval, self.bg)
                self.avgReady.emit(self)
                self.avg_count = 1

        if self.reg_f > 0:
            self.reg_avg = avg(self.aval, self.reg_s, self.reg_f)

        self.valueMeasured.emit(self)
        self.valueChanged.emit(self)

    cpdef void setAveraging(self, int n_avg):
        if n_avg == self.n_avg or n_avg < 1:
            return
        self.n_avg = n_avg
        if n_avg == 1:
            self.avg_aval = None
        if n_avg > 1:
            self.resize_avg_val(self.nelems)

    cpdef void getBg(self, bg_anum):
        self.bg = array.array('d')
        array.resize(self.bg, self.nelems)
        array.zero(self.bg)
        self.bg_count = bg_anum
        self.bg_anum = bg_anum
        self.bg_ready = False

    cpdef void setBgApply(self, bint state):
        self.bg_apply = state
        if state:
            self.bg_aval = array.array('d')
            array.resize(self.bg_aval, self.nelems)
            if using_numpy:
                self.bg_val = np.asarray(self.bg_aval)
        else:
            self.bg_aval = None

    cpdef void dropBg(self):
        self.bg = None
        self.bg_ready = False
        self.bg_count = -1

    cpdef void setRegProc(self, reg_s, reg_f):
        self.reg_s = reg_s
        self.reg_f = reg_f








from .misc_types cimport *


cdef extern from "cx.h" nogil:
    enum:
        CX_MAX_SERVER

    enum:
        CX_NULL_CHANID

    ctypedef int32  cxid_t
    ctypedef uint32 rflags_t
    ctypedef uint32 cxdtype_t

    ctypedef struct cx_time_t:
        int sec
        int nsec

    enum:
        __UNUSED__CX_TIME_SEC_TRUSTED
        CX_TIME_SEC_NEVER_READ

    ctypedef union CxAnyVal_t:
        float64  f64
        float32  f32
        int64    i64
        uint64   u64
        uint32   u32
        int32    i32
        int16    i16
        uint16   u16
        int8     i8
        uint8    u8
        char8    c8
        char32   c32


    enum:
        CXDTYPE_SIZE_MASK
        CXDTYPE_REPR_MASK
        CXDTYPE_REPR_SHIFT
        CXDTYPE_USGN_MASK

        #/* Representations */
        CXDTYPE_REPR_UNKNOWN
        CXDTYPE_REPR_INT
        CXDTYPE_REPR_FLOAT
        CXDTYPE_REPR_TEXT

    enum:
        CXDTYPE_UNKNOWN
        CXDTYPE_INT8
        CXDTYPE_INT16
        CXDTYPE_INT32
        CXDTYPE_INT64
        CXDTYPE_UINT8
        CXDTYPE_UINT16
        CXDTYPE_UINT32
        CXDTYPE_UINT64
        CXDTYPE_SINGLE
        CXDTYPE_DOUBLE
        CXDTYPE_TEXT
        CXDTYPE_UCTEXT

    size_t sizeof_cxdtype(cxdtype_t t)
    int    reprof_cxdtype(cxdtype_t t)


    enum: # flags
        CXRF_SERVER_MASK
        CXRF_CLIENT_MASK

        #/* Server-side flags -- CXRF_* */
        CXRF_CAMAC_NO_X       #// No 'X' from device
        CXRF_CAMAC_NO_Q       #// No 'Q' from device

        CXRF_IO_TIMEOUT       #// I/O timeout expired
        CXRF_REM_C_PROBL      #// Remote controller problem

        CXRF_OVERLOAD         #// Input channel overload

        CXRF_UNSUPPORTED      #// Unsupported feature/channel

        CXRF_INVAL            #// Invalid parameter
        CXRF_WRONG_DEV        #// Wrong device
        CXRF_CFG_PROBL        #// Configuration problem
        CXRF_DRV_PROBL        #// Driver internal problem
        CXRF_NO_DRV           #// Driver loading problem
        CXRF_OFFLINE          #// Device is offline

        CXRF_SERVER_HWERR_MASK
        CXRF_SERVER_SFERR_MASK

        #/* Client-side flags -- CXCF_* */
        #/* cda -- upwards from 16 */
        CXCF_FLAG_CALCERR     #// Formula calculation error
        CXCF_FLAG_DEFUNCT     #// Defunct channel
        CXCF_FLAG_OTHEROP     #// Other operator is active
        CXCF_FLAG_PRGLYCHG    #// Channel was programmatically changed
        CXCF_FLAG_NOTFOUND    #// Channel not found
        CXCF_FLAG_COLOR_WEIRD #// Value is weird

        CXCF_FLAG_CDA_MASK

        #/* Cdr -- downwards from 31 */
        CXCF_FLAG_ALARM_ALARM  #// Alarm!
        CXCF_FLAG_ALARM_RELAX  #// Relaxing after alarm
        CXCF_FLAG_COLOR_RED    #// Value in red zone
        CXCF_FLAG_COLOR_YELLOW #// Value in yellow zone

        CXCF_FLAG_CDR_MASK

        #/* General classification */
        CXCF_FLAG_HWERR_MASK
        CXCF_FLAG_SFERR_MASK
        CXCF_FLAG_SYSERR_MASK
        CXCF_FLAG_ALARM_MASK
        CXCF_FLAG_COLOR_MASK
        CXCF_FLAG_4WRONLY_MASK

    enum:
        CX_VALUE_LIT_MASK
        CX_VALUE_DISABLED_MASK
        CX_VALUE_COMMAND

    enum:
        CX_LOCK_RD
        CX_LOCK_WR
        CX_LOCK_ALLORNOTHING
        CX_LOCK_WRITE_SET
        CX_LOCK_WRITE_RLS





cdef class CdaObject:
    cdef:
        event **events
        int evnum

    def __cinit__(self, *args, **kvargs):
        self.events = NULL
        self.evnum = 0

    def __dealloc__(self):
        if self.evnum > 0:
            free(self.events)
            self.events = NULL
            self.evnum = 0

    cdef int add_event(self,int evmask, void *evproc, void *objptr, void *userptr):
        cdef:
            int x
            event *ev_t
            event *ev
            void *tmp

        if evmask == 0 or evproc == NULL:
            # event will not run
            return 0
        for x in range(self.evnum):
            ev_t = self.events[x]
            if ev_t.evmask == evmask and ev_t.evproc == evproc and \
            ev_t.objptr == objptr and ev_t.userptr == userptr:
                # event already exists
                return 0

        ev = <event*>malloc(sizeof(event))
        if not ev: raise MemoryError()

        ev.evmask, ev.evproc, ev.objptr, ev.userptr = evmask, evproc, objptr, userptr

        tmp = realloc(self.events, sizeof(event*) * (self.evnum + 1) )
        if not tmp: raise MemoryError()
        self.events = <event**>tmp
        self.events[self.evnum] = ev

        self.register_event(ev)
        self.evnum += 1
        return 1

    cdef int del_event(self, int evmask, void *evproc, void *objptr, void *userptr):
        cdef:
            event ev
            int ev_ind = -1
            int x

        for x in range(self.evnum):
            if self.events[x].evmask == evmask and self.events[x].evproc == evproc and \
            self.events[x].objptr == objptr and self.events[x].userptr == userptr:
                ev_ind = x
                break

        if ev_ind < 0: # event does not exist
            return 0
        self.unregister_event(self.events[ev_ind])
        if self.evnum == 1:
            free(self.events[ev_ind])
            # looks like self.events[ev_ind] = NULL - not needed here
            free(self.events)
            self.events = NULL
        else:
            memmove(&(self.events[ev_ind]), &(self.events[self.evnum-1]), sizeof(event*))
            tmp = realloc(self.events, sizeof(event*) * (self.evnum - 1))
            if not tmp: raise MemoryError()
            self.events = <event**>tmp
        self.evnum -= 1
        return 1

    cdef void register_event(self, event *ev):
        return

    cdef void unregister_event(self, event *ev):
        return



# textual include of basic level cda classes
include 'cxdtype.pxi'
include 'rflags.pxi'
include 'event.pxi'
include 'cdaobject.pxi'
include 'context.pxi'

cdef Context default_context=Context()

include 'basechan.pxi'
include 'dchan.pxi'
include 'ichan.pxi'
include 'chan.pxi'
include 'vchan.pxi'
include 'vpchan.pxi'
include 'strchan.pxi'
include 'chan_factory.pxi'

include 'cda_all.pxi'



from cx4.cxscheduler cimport sl_tout_proc, sl_enq_tout_after, sl_tid_t, sl_deq_tout

cdef void sltimer_proc(int uniq, void *privptr1, sl_tid_t tid, void *privptr2) with gil:
    cdef Timer t = <Timer>privptr2
    if t.repeat:
        t.tid = sl_enq_tout_after(0, NULL, t.usec, sltimer_proc, privptr2)
    else:
        t.active = 0
    t.timeout.emit()


cdef class Timer:
    """
    Timer class provides high-level to user in CX scheduler environment
    """
    cdef readonly:
        int usec
        int repeat
        int active
        sl_tid_t tid
        Signal timeout

    def __cinit__(self, int msec=1000):
        self.usec = msec * 1000
        self.repeat = 0
        self.active = 0
        self.timeout = Signal(object)

    cpdef void stop(self):
        """Stop any timer action."""
        if self.active:
            self.active = 0
            self.repeat = 0
            sl_deq_tout(self.tid)

    cpdef void start(self):
        """start timer for periodic shots.
         If timer is active it will be sopped and restarted"""
        if self.active == 1:
            sl_deq_tout(self.tid)
        self.active = 1
        self.repeat = 1
        self.tid = sl_enq_tout_after(0, NULL, self.usec, sltimer_proc, <void*>self)

    cpdef singleShot(self, int msec=0, proc=None):
        if msec > 0:
            self.setInterval(msec)
        if self.active == 1:
            sl_deq_tout(self.tid)
            self.tid = sl_enq_tout_after(0, NULL, self.usec, sltimer_proc, <void*>self)
        if proc is not None:
            self.timeout.connect(proc)
        self.active = 1
        self.repeat = 0
        self.tid = sl_enq_tout_after(0, NULL, self.usec, sltimer_proc, <void*>self)

    cpdef int interval(self):
        return <int>(self.usec/1000)

    cpdef void setInterval(self, int msec):
        """Set interval and restart timer if it is not running"""
        self.usec = 1000 * msec
        if self.active == 1:
            sl_deq_tout(self.tid)
            self.tid = sl_enq_tout_after(0, NULL, self.usec, sltimer_proc, <void*>self)




# chan to transfer python string
cdef class StrChan(BaseChan):
    cdef:
        readonly str val, prev_val
        char *cval

    def __cinit__(self, str name, **kwargs):
        self.cval = <char*>malloc(self.max_nelems)
        if not self.cval: raise MemoryError()
        self.val,self.prev_val = '', ''

    def __dealloc__(self):
        if not self.cval:
            free(self.cval)
            self.cval = NULL

    cdef void cb(self):
        self.prev_val = self.val
        c_len = self.get_data(0, self.max_nelems, <void*>self.cval)
        self.val = (<bytes>self.cval[:c_len]).decode('UTF-8')
        self.valueMeasured.emit(self)
        if self.val != self.prev_val:
            self.valueChanged.emit(self)

    cpdef void setValue(self, str value):
        bv = value.encode('UTF-8')
        cdef char *v = bv
        self.snd_data(CXDTYPE_TEXT, len(bv), <void*>v)









# scalar double channel
cdef class DChan(BaseChan):
    """
    One-double channel

    val - double last-known value
    prev_val - previous value
    tolerance - tolerance of emitting valueChanged signal
    """
    cdef:
        # all general properties defined in base classes
        readonly double val, prev_val, tolerance

    def __cinit__(self, str name, **kwargs):
        self.val, self.prev_val, self.tolerance = 0.0, 0.0, 0.0

    cdef void cb(self):
        self.prev_val = self.val
        self.check_exception(cda_get_dcval(self.ref, &self.val))
        if abs(self.val - self.prev_val) > self.tolerance or self.first_cycle:
            self.valueChanged.emit(self)
        self.valueMeasured.emit(self)
        self.first_cycle = False

    cpdef void setValue(self, double value):
        self.check_exception( cda_set_dcval(self.ref, value) )

    cpdef void setTolerance(self, double new_tolerance):
        self.tolerance = new_tolerance



# scalar int32 channel with simplified interface
cdef class IChan(BaseChan):
    cdef:
        # all general properties defined in base classes
        readonly int val, prev_val, tolerance

    def __cinit__(self, str name, **kwargs):
        #kwargs['dtype'] = CXDTYPE_UINT32
        #BaseChan.__init__(self, name, **kwargs)
        self.tolerance = 0

    cdef void cb(self):
        self.prev_val = self.val
        self.check_exception(cda_get_icval(self.ref, &self.val))
        if abs(self.val - self.prev_val) > self.tolerance or self.first_cycle:
            self.valueChanged.emit(self)
        self.valueMeasured.emit(self)
        self.first_cycle = False

    cpdef void setValue(self, int value):
        self.check_exception( cda_set_icval(self.ref, value) )

    cpdef void setTolerance(self, int new_tolerance):
        self.tolerance = new_tolerance



# Cython wrapper for cxv4 cda
# misc types declarations from bolkhov's misc_types.h
# by Fedor Emanov

cdef extern from "misc_types.h":
    ctypedef   signed int    int32
    ctypedef unsigned int    uint32
    ctypedef   signed short  int16
    ctypedef unsigned short  uint16
    ctypedef   signed char   int8
    ctypedef unsigned char   uint8

    ctypedef unsigned char   char8
    ctypedef        uint32   char32

    ctypedef   signed long long int int64
    ctypedef unsigned long long int uint64

    ctypedef float   float32
    ctypedef double  float64



from setuptools import setup
from setuptools.extension import Extension

import sys
import os.path
from pycx4.aux import cx_installpath
from pycx4.version import __version__

cxdir = cx_installpath()
if cxdir is None:
    print('Error: Unable to locate CX')
    sys.exit(1)
else:
    print('Using cx installed in:', cxdir)

cx_include = cxdir + '/include'
cx_lib = cxdir + '/lib'

USE_CYTHON = False
ext = '.c'
if os.path.isfile('./pycx4/pycda.pyx'):
    USE_CYTHON = True
    ext = '.pyx'

compiler_args = []
link_args = []

extensions = [
    Extension('pycx4.pycda', ['./pycx4/pycda' + ext],
              include_dirs=[cx_include],
              libraries=['cda', 'cx_async', 'useful', 'misc', 'cxscheduler'],
              library_dirs=[cx_lib],
              extra_compile_args=compiler_args,
              extra_link_args=link_args,
              )]

try:
    import PyQt4

    extensions.append(Extension('pycx4.q4cda', ['./pycx4/q4cda' + ext],
          include_dirs=[cx_include, '/usr/include/x86_64-linux-gnu/qt4/QtCore',],
          libraries=['cda', 'cx_async', 'useful', 'misc', 'Qt4cxscheduler', 'QtCore'],
          library_dirs=[cx_lib],
          extra_compile_args=compiler_args,
          extra_link_args=link_args,
    ))
except ImportError:
    print('PyQt4 not found. Building of corresponding module skipped')

try:
    import PyQt5

    extensions.append(Extension('pycx4.q5cda', ['./pycx4/q5cda' + ext],
              include_dirs=[cx_include, '/usr/include/x86_64-linux-gnu/qt5/QtCore'],
              libraries=['cda', 'cx_async', 'useful', 'misc', 'Qt5cxscheduler', 'Qt5Core'],
              library_dirs=[cx_lib],
              extra_compile_args=compiler_args,
              extra_link_args=link_args,
    ))
except ImportError:
    print('PyQt4 not found. Building of corresponding module skipped')


# Cython directives
directives = {
    'language_level': 3,
    'profile':   False,
    'linetrace': False,
    'c_string_type': 'bytes',
    'c_string_encoding': 'ascii',
    'boundscheck': False,
    'wraparound': False,
    'cdivision': True,
    'always_allow_keywords': False,
    'initializedcheck': False,
    'infer_types': True
}

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions, compiler_directives=directives)

setup(
    name='pycx4',
    version=__version__,
    url='https://github.com/femanov/pycx4/wiki',
    download_url='https://github.com/femanov/pycx4',
    author='Fedor Emanov',
    author_email='femanov@gmail.com',
    license='GPL',
    description='CXv4 control system framework Python bindings',
    long_description='CXv4 control system framework Python bindings, pycda and qcda modules',
    packages=['pycx4'],
    platforms='Linux',
    classifiers=[
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Cython",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],

    ext_modules=extensions
)




from cx4.cxscheduler cimport sl_add_fd, sl_del_fd, sl_fdh_t, sl_set_fd_mask

cdef void fd_event_proc(int uniq, void *privptr1, sl_fdh_t fdh, int fd, int mask, void *privptr2) with gil:
    cdef FdEvent fev = <FdEvent>privptr1
    fev.ready.emit(fev)


cdef class FdEvent:
    """
    File descriptor event class

    file - something with

    mask constants:
    SL_RD - Watch when descriptor is ready for read
    SL_WR - Watch when descriptor is ready for write
    SL_EX - Watch descriptor for exceptions (in fact -- OOB data)
    SL_CE - Watch for Connect Errors.  That's a hint for Xh_cxscheduler exclusively

    """
    cdef readonly:
        object file, ready
        int mask
        sl_fdh_t fhd # event id

    def __cinit__(self, file, int mask=SL_RD):
        self.file = file
        self.mask = mask
        self.ready = Signal(object)
        self.fhd = sl_add_fd(0, <void*>self, file.fileno(), mask, fd_event_proc, NULL)

    def __dealloc__(self):
        sl_del_fd(self.fhd)

    def set_mask(self, int mask):
        sl_set_fd_mask(self.fdh, mask)
        self.mask = mask
        # 2DO: need to think about exception-checking code here







cdef class SignalWakeupHandler(QAbstractSocket):
    signalReceived = QtCore.pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(QAbstractSocket.UdpSocket, parent)
        self.old_fd = None
        # Create a socket pair
        self.wsock, self.rsock = socket.socketpair(type=socket.SOCK_DGRAM)
        # Let Qt listen on the one end
        self.setSocketDescriptor(self.rsock.fileno())
        # And let Python write on the other end
        self.old_fd = signal.set_wakeup_fd(self.wsock.fileno())
        # First Python code executed gets any exception from
        # the signal handler, so add a dummy handler first
        self.readyRead.connect(lambda : None)
        # Second handler does the real handling
        self.readyRead.connect(self._readSignal)

    def __del__(self):
        # Restore any old handler on deletion
        if self.old_fd is not None and signal and signal.set_wakeup_fd:
            signal.set_wakeup_fd(self.old_fd)

    def _readSignal(self):
        # Read the written byte.
        # Note: readyRead is blocked from occuring again until readData()
        # was called, so call it, even if you don't need the value.
        data = self.readData(1)
        # Emit a Qt signal for convenience
        self.signalReceived.emit(data[0])





from .misc_types cimport *

cdef extern from "cx_common_types.h" nogil:
    ctypedef unsigned long         CxPixel

    ctypedef int CxDataRef_t
    enum:
        NULL_CxDataRef

    ctypedef struct CxKnobParam_t:
        char   *ident
        char   *label
        int     readonly
        int     modified
        double  value
        double  minalwd
        double  maxalwd
        double  rsrvd_d




# simple cycthon implementation of signals
from cpython cimport Py_INCREF,Py_DECREF

@cython.freelist(5)
cdef class Signal:
    cdef:
        void **callbacks
        int cnum

    def __cinit__(self, *args):
        self.callbacks = NULL
        self.cnum = 0

    def __dealloc__(self):
        cdef int ind
        for ind in range(self.cnum):
            Py_DECREF(<object>(self.callbacks[ind]))
        free(self.callbacks)

    cpdef connect(self, slot):
        cdef:
            void *tmp
            int ind
        if isinstance(slot, Signal):
            callback = slot.emit
        elif callable(slot):
            callback = slot
        else:
            raise Exception('A function was expected')
        for ind in range(self.cnum):
            if callback == <object>self.callbacks[ind]:
                return
        tmp = realloc(<void*>self.callbacks, sizeof(void*) * (self.cnum+1))
        if not tmp: raise MemoryError()
        self.callbacks = <void**>tmp
        self.callbacks[self.cnum] = <void*>callback
        self.cnum += 1
        Py_INCREF(callback)

    cpdef disconnect(self, callback):
        cdef:
            void *tmp
            int ind
        for ind in range(self.cnum):
            if callback == <object>self.callbacks[ind]:
                if self.cnum == 1:
                    free(self.callbacks)
                else:
                    memmove(&(self.callbacks[ind]), &(self.callbacks[self.cnum-1]), sizeof(void*))
                    tmp = realloc(self.callbacks, sizeof(void*) * (self.cnum-1))
                    if not tmp: raise MemoryError()
                    self.callbacks = <void**>tmp
                self.cnum -= 1
        Py_DECREF(callback)

    def emit(self, *args):
        cdef int ind
        for ind in range(self.cnum):
            (<object>(self.callbacks[ind]))(*args)

